/*
** Name: GP32 Yeti3D Port
** Desc: Portable 3D Engine based on Yeti3D
** Auth of the port : Gazin Matthieu <matthieu.gazin@laposte.net>
** Original Auth of the engine: Derek J. Evans <derek@theteahouse.com.au>
** Copyright (C) 2003 Derek J. Evans. All Rights Reserved.
**
** YY  YY EEEEEE TTTTTT IIIIII 33333  DDDDD
** YY  YY EE       TT     II       33 DD  DD
**  YYYY  EEEE     TT     II     333  DD  DD
**   YY   EE       TT     II       33 DD  DD
**   YY   EEEEEE   TT   IIIIII 33333  DDDDD
**
**             OOOOO    NN    NN
**            OO   OO   NNNN  NN
**            OO   OO   NN NN NN 
**            OO   OO   NN  NNNN
**             OOOOO    NN    NN
**
**  GGGGGGGG     PPPPPPP    333333     222222222     
** GGGGGGGGGG   PPPPPPPPP    333333   22222222222
** GG           PP     PP        33   22       22
** GG     GG    PPPPPPPPP     33333         2222
** GG    GGGG   PPPPPPPP      33333      222    
** GG      GG   PP               33    222 
** GGGGGGGGGG   PP           333333   22222222222
**  GGGGGGGG    PP          333333    22222222222
**
**
** ----------------------------------------------------
**
** V 0.0.3.0 to V 0.0.3.7
**       Added Converting to direct acces the video
**       Added CPU mode added for enable the cache
**       Added GP32 ID check for private beta added
**       Added Direct control input check added
**       Added MD2 model render
**       Change Redefine the RGB basic buffer format
**       Added GP32 Profile code for Yeti
**       Change Following Dereck for global code change (ex:YETIROM)
**       Added Entity can be draw by sprite or MD2 model now
**       Added ASM DIVISE and FIXDIVIDE function for ARM.
**       Added a little ligne/console write for init engine stats
**       Added convertion LUA for speed up the draw part
**       Added ASM copy screen fonction
**       Added cleanig cell draw buffer only if draw overdistance for don't loose time.
**       Added external map loader
**       Added external raw texture loader
**       Added external pallette loader
**       Change Desactivated fire and entity
**       Change Cleaned GP32 code and Yeti Code.
**       Change Lasy to comment yeti 3D modified file, it will be do later :p
**       Change Bit change in classique C function (copy screen) for optimise a little
**       Change Code change +1/2 fps on large zone (test on wipeout level by orion)
**       Remove All asm part remove so i remove my ASM DIVISE and FIXDIVIDE too.
**       Added external loader sprite
**       ...... (more but don't remember all :p)
** 
** V 0.0.2.9
**       Added Structure code added using reference file for direct acces hardware of the gp32
**
** V 0.0.2.8
**       Change Using Direct acces LCD
**
** V 0.0.1.1 to V 0.0.2.7  
**       Added Using SDK no special optimisation for the GP32
**       Added Try to port the engine to the GP32.
**
*/


// Basic include from the base sdk /* not really use, only for IO */
#include "gpdef.h"
#include "gpstdlib.h"
#include "gpgraphic.h"
#include "gpfont.h"
#include "gpstdio.h"
#include "gpmem.h"
#include "gpmm.h"
//#include <stdlib.h>


// Include from Yeti 3d engine
#include "yeti.h"
#include "viewports.h"
#include "game.h"


// Include for GP32 /* That is GOOD !!!! ^_^ */
#include "gfxlib.h"
#include "bioslib.h"
#include "gp32.h"
#include "gp32_font.h"


// Declare Yeti 3D structure engine and viewport
framebuffer_t framebuffer;
yeti_t yeti;
rom_map_t rommap;

YETI_ROM texture_t textures[YETI_TEXTURE_MAX];
palette_t palette;
YETI_ROM lua_t lua;

// not use in demo (function are in the code but not use too)
//#define SPRITE_MAX 5
//YETI_ROM sprite_t sprites[];
//YETI_ROM u16 spr_00[];
//YETI_ROM u16 spr_01[];
//YETI_ROM u16 spr_02[];
//YETI_ROM u16 spr_03[];
//YETI_ROM u16 spr_ball1[];


// GP32 port info !!!
#define YETI_GP32_PORT_STR_TITLE      "Yeti3D Engine -> GP32 PORT"
#define YETI_GP32_PORT_VERSION        "0.0.3.7"
#define YETI_GP32_PORT_STR_AUTOR      "GAZIN Matthieu"
#define YETI_GP32_PORT_STR_CAPTION    (YETI_GP32_PORT_STR_TITLE " " YETI_GP32_PORT_STR_VERSION " - " YETI_GP32_PORT_STR_AUTOR)


int key_pressed;         // For input check
int byebye = 0;            // on quitte ;)

// Compile under ADS
#define GP32_ADS 

#ifdef GP32_ADS
#define ONE_SECOND 2000 
#else
#define ONE_SECOND 1000
#endif


#ifdef GP32_ADS
int My_GpKeyGet(void)
{
int value=0;

#define rKEY_A          0x4000
#define rKEY_B          0x2000
#define rKEY_L          0x1000
#define rKEY_R          0x8000
#define rKEY_UP         0x0800
#define rKEY_DOWN       0x0200
#define rKEY_LEFT       0x0100
#define rKEY_RIGHT      0x0400
#define rKEY_START      0x0040
#define rKEY_SELECT     0x0080
#define rPBDAT          (*(volatile unsigned *)0x1560000c)
#define rPEDAT          (*(volatile unsigned *)0x15600030)

unsigned long gpb = rPBDAT;   // 0x156
unsigned long gpe = rPEDAT;

if ((gpb & rKEY_LEFT) == 0) value|=GPC_VK_LEFT;
if ((gpb & rKEY_RIGHT) == 0) value|=GPC_VK_RIGHT;
if ((gpb & rKEY_UP) == 0) value|=GPC_VK_UP;
if ((gpb & rKEY_DOWN) == 0) value|=GPC_VK_DOWN;
if ((gpb & rKEY_A) == 0) value|=GPC_VK_FA;
if ((gpb & rKEY_B) == 0) value|=GPC_VK_FB;
if ((gpb & rKEY_L) == 0) value|=GPC_VK_FL;
if ((gpb & rKEY_R) == 0) value|=GPC_VK_FR;
if ((gpe & rKEY_SELECT) == 0) value|=GPC_VK_SELECT;
if ((gpe & rKEY_START) == 0) value|=GPC_VK_START;

return value;
}

 // redefine the GpKeyGet
 #define GpKeyGet My_GpKeyGet
#endif

// Classic delay
#define DelayA(); while(!(GpKeyGet()&GPC_VK_FA)); while(GpKeyGet()&GPC_VK_FA);
#define DelayB(); while(!(GpKeyGet()&GPC_VK_FB)); while(GpKeyGet()&GPC_VK_FB);


// some fps count
unsigned long fps;
unsigned long fps_time, fps_tick;//, fps_min, fps_max;
char fps_str[100];

// for 133 personal tick get
#define __gp32__timermul__ 100
#define __gp32__timerdiv__ 236
long gp32_GpTickCountGet(void);

// main boucle
void gp32_to_yeti3D(void);

///  loader de map de l'editeur
void gp32_load_map(char* filename);

// loader fichier extern
unsigned char *get_gpg(char *filename);
unsigned char *get_raw(char *filename);
unsigned char *get_pal(char *filename);
unsigned char *get_sprite(char *filename);
unsigned char *tempo_texture_cache;
void load_texture_and_pal_files(char *name);

// fonction de copie
//void fbitblt16(unsigned char *source, unsigned char *dest); 
//void fbitblt16_w(unsigned short *source, unsigned short *dest); 


// Amm for little console system
void gp32_GpTextOut(unsigned char *buffer,int X,int Y,char *A,int col,int bold);
#define Msg(X,Y,A,col) \
  gp32_GpTextOut(framebuffer.pixels,X,Y,A,col,0);
#define MsgBold(X,Y,A,col) \
  gp32_GpTextOut(framebuffer.pixels,X,Y,A,col,1);
int console_write_y = 1;
#define console_write(TEXT,COULEUR) \
 gp32_GpTextOut(currentbuffer,1,console_write_y,TEXT,COULEUR,0); \
 console_write_y+=11;
#define console_write_no_lineup(TEXT,COULEUR) gp32_GpTextOut(currentbuffer,1,console_write_y,TEXT,COULEUR,0);

 

// video
#define BUFFERSIZE 320*240*2*3
static unsigned char fbuf[BUFFERSIZE+4096]; 
struct LCDinfo lcd;
unsigned char *buffer;
unsigned char *vrambuffer,*currentbuffer;
extern void EnableCache(void);
unsigned long *GP32_PAL=(unsigned long *)0x14A00400; //32 bits?

float Gamma=1.5;
char tmp_buffer[255];


// special fonction polydraw asm pas encore au point 
// a finir
#ifdef __PATCH_DRAW_POLY__
 typedef struct 
 {
   int Sx, l, u, v;
 } edge_t; 

 edge_t EdgeLeft[200];
 edge_t EdgeRight[200];
#endif




// MAIN GP32 BOUCLE !!!!
void GpMain(void *arg)
{    
	ARMDisableInterruptSpiv();
	setCPUSpeed (132000000, 0x3a011, 3);
	buffer =  (unsigned char *)(((long)(fbuf+4095) & ~4095));

    // 2 ca suffira bien 
//	setup16bppMode(&lcd,buffer,3);
	setup16bppMode(&lcd,buffer,2);
	lcd.clkval = 8;
	lcd.posx = 0;
	lcd.posy = 0;
	lcd.x = 240;
	lcd.y = 320;
	lcd.offx = 0;
	lcd.offy = 0;
	setupLCDMode(&lcd,1);		
	//becoz it doesn't work well have to see this pb
	GpGraphicModeSet(16, NULL);  	

	initTripleBufferDMA(&lcd,DMA0_INT,NULL);
	cache_invalidate_all();
	mmuChange(buffer,buffer+BUFFERSIZE,CACHE_CNW);
	 //tripleBufferDMA(&lcd,0);		
	vrambuffer = doubleBuffer(&lcd,0);
	currentbuffer = vrambuffer;
	vrambuffer = doubleBuffer(&lcd,0);	
	EnableCache();


  console_write("---------------------------------------",COLOR_WHITE);
  console_write(YETI_GP32_PORT_STR_TITLE,COLOR_YELLOW);
  gp_str_func.sprintf(tmp_buffer,"Version : %s  by %s",YETI_GP32_PORT_VERSION,YETI_GP32_PORT_STR_AUTOR);
  console_write(tmp_buffer,COLOR_YELLOW);
  console_write("---------------------------------------",COLOR_WHITE);

  // Bon ca peut parraitre con vue ke on est deja la mais bon c pour le fun on va dire ke c'est bon quand meme :p
  console_write("Init triple DMA buffer OK",COLOR_GREEN);
  console_write("Enable CPU cache OK",COLOR_GREEN);
  console_write("CPU overcloack OK",COLOR_GREEN);

  console_write_no_lineup("Init Fat System",COLOR_ORANGE);
   GpFatInit();
  console_write("Init Fat System OK",COLOR_GREEN);
  
  console_write_no_lineup("Set Relative Path",COLOR_ORANGE);
   GpRelativePathSet("gp:\\gpmm\\maps\\"); 
  console_write("Set Relative Path OK",COLOR_GREEN);
  

 console_write_no_lineup("Yeti3D Engine init",COLOR_ORANGE);
  // init engine
  yeti_init(&yeti, &framebuffer, &framebuffer, textures, palette, lua);
 console_write("Yeti3D Engine init OK !!!",COLOR_GREEN);


 console_write_no_lineup("Loading Map",COLOR_ORANGE);
  gp32_load_map("e1m1.y3d");
  yeti_load_map(&yeti, &rommap);
 console_write("Loading Map OK !!!",COLOR_GREEN);

 console_write_no_lineup("Game Engine init and LUA Calculation",COLOR_ORANGE);
  game_init(&yeti);
 console_write("Game Engine init and LUA Calculation OK !!!",COLOR_GREEN);


 // chargement des textures
 load_texture_and_pal_files("e1m1");


 // Memory left after all ?
 gp_str_func.sprintf(tmp_buffer,"Memory available : %d kb",gm_availablesize()/1024);
 console_write(tmp_buffer,COLOR_GREEN);

 console_write("Running ....",COLOR_GREEN);


  while(!byebye)
  {
   // Loop of the gp32 port
   gp32_to_yeti3D();   
  }
  GpAppExit();
}


void gp32_to_yeti3D(void)
{
  static unsigned long Time;
// for C screen copie
  int x,y,y_sav;

  // for screen copie
  unsigned short *p_buf;


 // All this buggy tick function very bad grrrr !!!!!
 // i've to found a way for do all the time coorectly
 // loose time here, have to be redone !!!
/*
    fps++;
//    fps_tick = GpTickCountGet();
    fps_tick = gp32_GpTickCountGet();
   		
    if((int)(fps_tick - fps_time) >= 1000)
    {
     fps_time=fps_tick;
     gp_str_func.sprintf(fps_str,"FPS : %i",fps);
     fps=0;     
    }
	*/
  if ((int)(Time - GpTickCountGet()) < 0)
//  if (((int)Time - gp32_GpTickCountGet()) < 0)
  {
    Time = GpTickCountGet() + YETI_VIEWPORT_INTERVAL;
//    Time = gp32_GpTickCountGet() + YETI_VIEWPORT_INTERVAL;

    yeti.keyboard.up=yeti.keyboard.down=yeti.keyboard.left=yeti.keyboard.right=yeti.keyboard.a=yeti.keyboard.b=yeti.keyboard.l=yeti.keyboard.r=0;

    key_pressed=GpKeyGet(); // Get input data    

    // bye bye !!!
    if (key_pressed & GPC_VK_START)
    byebye = 1;

    if (key_pressed & GPC_VK_UP)
     yeti.keyboard.up     = 1;
    if (key_pressed & GPC_VK_DOWN)
     yeti.keyboard.down   = 1;
    if (key_pressed & GPC_VK_LEFT)
     yeti.keyboard.left   = 1;
    if (key_pressed & GPC_VK_RIGHT)
     yeti.keyboard.right  = 1;
    if (key_pressed & GPC_VK_FA)
     yeti.keyboard.a  = 1;
    if (key_pressed & GPC_VK_FB)
     yeti.keyboard.b  = 1;
    if (key_pressed & GPC_VK_FL)
     yeti.keyboard.l  = 1;
    if (key_pressed & GPC_VK_FR)
     yeti.keyboard.r  = 1;


    // link buffer draw to gp32 video buffer
 
 // version C refait pour optimiser le calcul de donner

 p_buf = (unsigned short *)currentbuffer;
 for ( y = 0; y < 240; y++)
  {
    y_sav= 239-y;
    for(x = 0; x<320; x++)
    {
     p_buf[(x*240)+y_sav] = framebuffer.pixels[y][x];
   }
  }
 

// Asm version disabled for gpl
//p_buf = (unsigned short *)currentbuffer;
//fbitblt16_w(framebuffer.pixels, p_buf);

// asm 2 pixel fast copy not working yet
//copyBuffer(LCD_SCREEN,framebuffer.pixels);
//p_buf = (unsigned short *)currentbuffer;
//copyBuffer(p_buf,framebuffer.pixels);
//copyBuffer(0x0c7b4000, framebuffer.pixels);

/* 
Personal note :
i've to redone the funtion in asm do use the 2 pixel copy trick in asm for speed up the copie.
*/

  // loose time here, have to be redone !!!
//  gp32_GpTextOut((unsigned char*)p_buf, 1,1,fps_str,COLOR_YELLOW,0);
  
  // Engine loop
  game_loop(&yeti);
  }
}
 

void gp32_Printc(unsigned long x,unsigned long y,char c,unsigned char *scr,int col,int bold)
{
  int indice=c-32; //index in gp32_font.h
  int indicex,indicey;  
  unsigned long i,j;
  int colB = ((((col>>1)&31)>>1)<<1)|((((col>>6)&31)>>1)<<6)|((((col>>11)&31)>>1)<<11)|1;
  

  if (indice>=96) indice=0;
  if (indice<0) indice=0;
  indicey=(indice>>4)*9;
  indicex=(indice&0xf)<<3;

  
  	if (bold)
  	{
		  for (i=0;i<8;i++)
	      for (j=0;j<9;j++) 
	      {
		    if (font[j+indicey][i+indicex]=='#') 
			{
				((unsigned char *)(scr))[(239-(y+j)+((i+x)*240))<<1]=col&0xFE;
				((unsigned char *)(scr))[(239-(y+j)+((i+x)*240))<<1|1]=col>>8;
			}			
			if (font[j+indicey][i+indicex]=='.')
			{
				((unsigned char *)(scr))[(239-(y+j)+((i+x)*240))<<1]=colB&0xFE;
				((unsigned char *)(scr))[(239-(y+j)+((i+x)*240))<<1|1]=colB>>8;
			}
		  }
	}
	else
	{
		  for (i=0;i<8;i++)
	      for (j=0;j<9;j++) 
		    if (font[j+indicey][i+indicex]=='#') 
			{
				((unsigned char *)(scr))[(239-(y+j)+((i+x)*240))<<1]=col&0xFE;
				((unsigned char *)(scr))[(239-(y+j)+((i+x)*240))<<1|1]=col>>8;
			}  	
	}
}


// gp_str_func.gpstrlen fait perdre bcp plus de temps que STRLEN
// le mieux est de faire les calcul de fps apres coups car on perd pas mal de temp et a afficher les fps
// <> entre gp_str_func.gpstrlen et strlen environ 1/2 fps
// y'a vraiment des trucs  la con sur gp32 c'est la 1er fois que je vois ca moi
void gp32_GpTextOut(unsigned char *buffer,int X,int Y,char *A,int col,int bold)
{
		int l;
	    int k;
		int msg_x,msg_y;  		
		if (A)
		{
			l=strlen(A);
//			l=gp_str_func.gpstrlen(A);
			msg_x=X;
			msg_y=Y;
			k=0;   
			while (k<l)
			{
				if (A[k]!='\n') 
				{
					gp32_Printc(msg_x,msg_y,A[k],buffer,col,bold);
					msg_x+=7;
				}	
				else 
				{
				    msg_y+=8;
					msg_x=X;
				}
				if (msg_x>=(320-7))
				{
					msg_y+=8;
					msg_x=0;
				}
				if (msg_y>=(240-8))
				{
					msg_x=0;
					msg_y=0;	      
				}
				k++;
			}
		}				
}


long gp32_GpTickCountGet(void)
{
	return (GpTickCountGet()*__gp32__timermul__/__gp32__timerdiv__);
}

 
void gp32_load_map(char* filename)
{
  // on suppose que le path vers gpmm/maps a ete mis
ERR_CODE IO_ERROR_FILE;
F_HANDLE IO_FILE;
unsigned long m_size;
 

   IO_ERROR_FILE = GpFileOpen(filename, OPEN_R, &IO_FILE); // ouverture en lecture
   if (IO_ERROR_FILE == SM_OK)	// donc on est bon
   {
    IO_ERROR_FILE = GpFileRead(IO_FILE, &rommap, sizeof(rommap), &m_size);

    if (IO_ERROR_FILE == SM_OK)
    {
      // rien a dire je suis une bete ca marche (m'enfin a ce niveau la ca devrait pas tre trop dur non plus)
      GpFileClose(IO_FILE);
    }
    else
    {  
      GpFileClose(IO_FILE);  
      // erreur ya une couille la 
     console_write_y+=11;
     console_write("SYSTEM: Cannot Load Map File !!!",COLOR_RED);
     while (!GpKeyGet());
     while (GpKeyGet());
     GpAppExit();
    }
   }
} 

/*
GPG: 
0 - 3 : ID de l'archive archivo. "gpg " (4byte)
4 - 7 : Date (4byte)
8 - X : Donnes
*/
unsigned char *get_gpg(char *filename)
{
	unsigned char	*data;
	F_HANDLE		hfile;
	unsigned long	p_read_count;
	unsigned long	fsize;
	hfile = NULL;
	if (SM_OK != GpFileOpen(filename, OPEN_R, &hfile)) 
	{ 
      // erreur ya une couille la 
      return NULL;
	}  
	GpFileGetSize(filename, &fsize);
	GpFileSeek(hfile, FROM_BEGIN, 8, 0); // on decale bien
	data = gm_zi_malloc(fsize); //  * sizeof(unsigned short)
	GpFileRead(hfile, data, fsize, &p_read_count);
	GpFileClose(hfile);
	return (data);
}


unsigned char *get_raw(char *filename)
{
	unsigned char	*data;
	F_HANDLE		hfile;
	unsigned long	p_read_count;
	unsigned long	fsize;
	hfile = NULL;
	if (SM_OK != GpFileOpen(filename, OPEN_R, &hfile)) 
	{ 
      // erreur ya une couille la 
      return NULL;
	}  
	GpFileGetSize(filename, &fsize);
	data = gm_zi_malloc(fsize);
	GpFileRead(hfile, data, fsize, &p_read_count);
	GpFileClose(hfile);
	return (data);
}


unsigned char *get_pal(char *filename)
{
	unsigned char	*data;
	F_HANDLE		hfile;
	unsigned long	p_read_count;
	unsigned long	fsize;
	hfile = NULL;
	if (SM_OK != GpFileOpen(filename, OPEN_R, &hfile)) 
	{ 
      // erreur ya une couille la 
      return NULL;
	}  
	GpFileGetSize(filename, &fsize);
	data = gm_zi_malloc(fsize);
	GpFileRead(hfile, data, fsize, &p_read_count);
	GpFileClose(hfile);
	return (data);
}

unsigned char *get_sprite(char *filename)
{
	unsigned char	*data;
	F_HANDLE		hfile;
	unsigned long	p_read_count;
	unsigned long	fsize;
	hfile = NULL;
	if (SM_OK != GpFileOpen(filename, OPEN_R, &hfile)) 
	{ 
      // erreur ya une couille la 
      return NULL;
	}  
	GpFileGetSize(filename, &fsize);
	data = gm_zi_malloc(fsize);
	GpFileRead(hfile, data, fsize, &p_read_count);
	GpFileClose(hfile);
	return (data);
}


int	lua_rgb (int r, int g, int b)
{
	int	r1 = (MIN (r, 255) * 31) / 255;
	int	g1 = (MIN (g, 255) * 31) / 255;
	int	b1 = (MIN (b, 255) * 31) / 255;

//	return (r1 << 10) + (g1 << 5) + b1;
return ( ( ( r1 ) << 11 )  |  ( ( g1 ) << 6 )  |  ( ( b1 ) << 1 ) );
}

// only indicate general file name extension raw and pal will be set
void load_texture_and_pal_files(char *name)
{
	unsigned char	*data;
	F_HANDLE		hfile;
	unsigned long	p_read_count;
	unsigned long	fsize;
    char filename[100];
    unsigned long num_tex;
	int a,i;
	hfile = NULL;



    console_write("Loading Data ...",COLOR_GREEN); 
    console_write_no_lineup("Loading pallette",COLOR_ORANGE);
    // pallette first
    gp_str_func.sprintf(filename,"%s.pal",name);
	if (SM_OK != GpFileOpen(filename, OPEN_R, &hfile)) 
	{ 
      // erreur ya une couille la 
     console_write_y+=11;
     console_write("SYSTEM: Cannot Load Pallette File !!!",COLOR_RED);
     while (!GpKeyGet());
     while (GpKeyGet());
     GpAppExit();
    }  
	GpFileGetSize(filename, &fsize);
	data = gm_zi_malloc(fsize);
	GpFileRead(hfile, data, fsize, &p_read_count);
	GpFileClose(hfile);
    console_write("Loading pallette OK !!!",COLOR_GREEN);

    // Fixed size
    gm_memcpy(palette,data,768);



    console_write_no_lineup("Loading Textures",COLOR_ORANGE);
    // texture next
    gp_str_func.sprintf(filename,"%s.raw",name);
	if (SM_OK != GpFileOpen(filename, OPEN_R, &hfile)) 
	{ 
      // erreur ya une couille la 
     console_write_y+=11;
     console_write("SYSTEM: Cannot Load Texture File !!!",COLOR_RED);
     while (!GpKeyGet());
     while (GpKeyGet());
     GpAppExit();
    }  
	GpFileGetSize(filename, &fsize);
	data = gm_zi_malloc(fsize);
	
	num_tex=fsize/(64*64);
	
	GpFileRead(hfile, data, fsize, &p_read_count);
	GpFileClose(hfile);

    gp_str_func.sprintf(filename,"Loading Textures (%i founds) OK !!!",num_tex);
    console_write(filename,COLOR_GREEN);

    // actually 64*64 textures
    gm_memcpy(textures,data,64*64*num_tex);
    
    

    // mise a jour LUA 
    for(a=0;a<64;a++)
    {
     for(i=0;i<256;i++) 
     {
	  lua[a][i]=lua_rgb(palette[i][0]*a/16,palette[i][1]*a/16,palette[i][2]*a/16); 
     }
    }

}


